#; ScaraOS bootloader by Gianni Tedesco <gianni@ecsc.co.uk>
#; Copyright (c) 2001 Gianni Tedesco
#; Released under the terms of the GNU GPL Version 2 or later

.code16
.text
.global _start

DEF_BOOT_SEG		= 0x07C0
INIT_BOOT_SEG		= 0x9000

GDT_CODE_SEG		= 16
GDT_DATA_SEG		= 24

_start:
	#; Boot sectors are (supposedly) meant to start with a jmp
	jmp start
	nop
start:
	#; Move ourselves out of the way
	movw    $DEF_BOOT_SEG, %ax
	movw    %ax, %ds
        movw    $INIT_BOOT_SEG, %ax
        movw    %ax, %es
        movw    $256, %cx
        subw    %si, %si
        subw    %di, %di
        cld
        rep
        movsw
	ljmp	$INIT_BOOT_SEG, $init_seg

init_seg:
	call setup_iseg

	mov	$lmsg, %si
	call	print_str

	#; Best to have interrupts disabled
	#; while we are doing pmode jigggery pokery,
	#; we dont reenable until we are in the kernel
	#; and have an IDT loaded...
	cli

	#; Enable Gate A20
	#; Some laptops will b0rk here cos we don't wait
	call	enable_a20

	#; Enter pmode for the first time
	call	enter_pmode

	#; Back to real mode
	dec	%cx
	mov	%ecx, %CR0
	call	setup_iseg
	sti

	#; Print message
	mov	$kmsg, %si
	call	print_str

	#; TODO: Load kernel from next sectors
	#; TODO: Decode ELF binary in to linear

	#; Jump to 32bit code to run kernel
	cli
	call 	enter_pmode

	#; Jump in to the 32 bit code
	mov	$GDT_CODE_SEG, %cx
	jmp 	%cs:c32s

.code32
c32s:
	jmp	c32s
	#;TODO: Jump in to kernel
.code16

#;-- Code to enter 32 bit pmode
enter_pmode:
	#; Load in our GDT
	movw	$32, (gdt_data+2)	#; Fill in GDT length
	lgdt	gdt_data

	#; Go to 32bit protected mode
	mov	%CR0, %ecx
	inc	%cx
	mov	%ecx, %CR0

	#; Setup our segment registers
	mov	$GDT_DATA_SEG, %ax	#; segment selector
	mov	%ax, %ds
	mov	%ax, %es
	mov	%ax, %fs
	mov	%ax, %gs

	ret

#;-- Print a string (%si is str ptr)
print_str:
wstr_1:	lodsb
	and	%al, %al
	jz	ret_p
	mov	$0x0E, %ah
	mov	$0x0007, %bx
	int	$0x10
	jmp	wstr_1
ret_p:	ret

#;-- Setup our segment registers
setup_iseg:
	movw	$0x4000,%di
	movw	%ax, %ds
	movw 	%ax, %ss
	movw	%di, %sp	#; Stack is at INIT_BOOT_SEG:0x4000
	movw	$INIT_BOOT_SEG, %ax
	movw	%ax, %es
	ret

#;-- Code to enable Gate A20
empty_8042:
	pushl	%ecx
	movl	$100000, %ecx
	
empty_8042_loop:
	decl	%ecx
	jz	empty_8042_end_loop

	call	delay

	inb	$0x64, %al	
	testb	$1, %al
	jz	no_output

	call	delay
	inb	$0x60, %al
	jmp	empty_8042_loop

no_output:
	testb	$2, %al
	jnz	empty_8042_loop
empty_8042_end_loop:
	popl	%ecx
	ret

delay:
	outb	%al, $0x80
	ret

enable_a20:
	call	empty_8042

	movb    $0xD1, %al	#; command write
	outb    %al, $0x64
	call    empty_8042

	movb    $0xDF, %al	#; A20 on
	outb    %al, $0x60
	call    empty_8042
	
	inb	$0x92, %al
	orb	$02, %al
	outb	%al, $0x92
	
	ret

#;-- DATA

#; JMPI 0x1000:$GDT_CODE_SEG
code32_start:
	.long	0x1000
inst:
	.byte	0x66, 0xea
code32:	.long	0x1000
	.word	GDT_CODE_SEG

#;-- BOOT MESSAGE
lmsg:	.ascii "Loading ScaraOS..."
	.byte 13, 10, 0

#;-- ELF ERROR
kmsg:	.ascii "PMODE OK! Loading kernel"
	.byte 13, 10, 0

#;-- Misc. Configuration
sec_per_track:		.byte 18
total_heads:		.word 2
boot_drv:		.word 0
kern_siz:		.word 51

#;-- GDT Info
gdt_data:
	.word	0x8000
	.word	0
gdt_table:
	#; Null and dummy descriptors
	.word	0, 0, 0, 0
	.word	0, 0, 0, 0

	#; Code segment
	.word	0xFFFF		#; 4Gb
	.word	0		#; base address = 0
	.word	0x9A00		#; code read/exec
	.word	0x00CF		#; Granularity = 4096, 386

	#; Data segment
	.word	0xFFFF		#; 4Gb
	.word	0		#; base address = 0
	.word	0x9200		#; code read/write
	.word	0x00CF		#; Granularity = 4096, 386


#;-- BOOT IDENTIFIER
.org 510
boot_flag:	.word 0xAA55
