#include <kernel.h>
#include <arch/dma.h>
#include <arch/fdc.h>
#include <arch/8259a.h>
#include <arch/idt.h>
#include <fs/vfs.h>

#define _FDC_PREFIX "  (drv/fdc): "

void *tb;

volatile int	done_irq=0;	/* No of irqs recieved */
u_int32_t 	fdc_timeout=0;	/* Timeout time */
int8_t		sr0=0;		/* Status register */
int8_t		fdc_track=0xff;	/* Track number */

int8_t		status[7]={0,0,0,0,0,0};
int8_t		statsz = 0;

/* Supported drive geometries */
fdc_geom_t geometries[]={
	{2, 80, 18, 0x54, 0x1b},	/* 1.44M floppy */
	{2, 80, 21, 0x0c, 0x1c},	/* 1.68M floppy */
};
p_fdc_geom geometry=&geometries[0];


/* INTERRUPT SERVICE ROUTINE */
void fdc_isr(void)
{
	done_irq++;
	irq_eoi(6);
}
void _irq6(void);
asm(
	".globl _irq6		\n"
	"_irq6:			\n"
	"	pusha		\n"
	"	call fdc_isr	\n"
	"	popa		\n"
	"	iret		\n"
);

void fdc_linear2hts(u_int32_t block, int *head, int *track, int *sector)
{
	*head = (block % (geometry->spt * geometry->heads)) / (geometry->spt);
	*track = block / (geometry->spt * geometry->heads);
	*sector = block % geometry->spt + 1;
}

/* Wait for floppy command completion */
int fdc_wait(int sensei)
{
	FDC_SET_TIMEOUT(100);
	
	while ( done_irq == 0 && FDC_GOT_TIME ) {
		/* kprint("IRQ=%u NOW=%u TIMEOUT=%u\n",
		* 	done_irq, ticks, fdc_timeout); */
		sleep_ms(10);
	}

	/* Read command results */
	statsz=0;
	while( (statsz<7) && (inb(FDC_MSR) & (1<<4))  ){
		status[statsz++]=fdc_getbyte();
	}
	
	/* Sense interrupt */
	if ( sensei ) {
		fdc_sendbyte(CMD_SENSEI);
		sr0=fdc_getbyte();
		fdc_track=fdc_getbyte();
	}

	done_irq=0;

	if (!FDC_GOT_TIME) {
		if (inb(FDC_DIR) & 0x80) {
			kprint("[fdc]: Disk Changed!!\n");
			return 1;
		}else{
			return 0;
		}
	}else{
		return 1;
	}
}

void fdc_sendbyte(int8_t b)
{
	volatile int msr;
	int tmo;

	for (tmo=0; tmo<128; tmo++) {
		msr = inb(FDC_MSR);
		if ((msr & 0xc0) == 0x80) {
			outb(FDC_DATA,b);
			return;
		}
		inb(0x80);   /* delay */
	}
}

int8_t fdc_getbyte(void)
{
	volatile int msr;
	int tmo;

	for (tmo=0; tmo<128; tmo++) {
		msr = inb(FDC_MSR);
		if ((msr & 0xd0) == 0xd0) {
			return inb(FDC_DATA);
		}
		inb(0x80);   /* delay */
	}
	return -1;
}

int fdc_seek(int8_t track)
{
	/* Already there */
	if ( fdc_track == track ) return 1;

	fdc_sendbyte(CMD_SEEK);
	fdc_sendbyte(0);
	fdc_sendbyte(track);

	if ( !fdc_wait(1) ) {
		return 0;
	}

	/* Wait for head to settle */
	sleep_ms(15);

	if ( (sr0!=0x20) || (fdc_track != track) ) return 0;

	return 1;
}

void fdc_recalibrate(void)
{
	fdc_sendbyte(CMD_RECALIBRATE);
	fdc_sendbyte(0);

	fdc_wait(1);
}

void fdc_reset(void)
{
	/* Stop the lot ! */
	outb(FDC_DOR, 0);

	/* Data rate select 500k/s */
	outb(FDC_CCR, 0);
	
	/* Re start */
	outb(FDC_DOR, 0x1c);
	
	if ( done_irq ) {
		kprint(_FDC_PREFIX "Resetting floppy caused interrupt\n");
		fdc_wait(1);
	}
	
	/* Specify device timings */
	fdc_sendbyte(CMD_SPECIFY);
	fdc_sendbyte(0xdf);
	fdc_sendbyte(0x02);

	fdc_seek(1);
	fdc_recalibrate();
}

void __init fdc_init(void)
{
	tb=kmalloc(GFP_DMA, 1024);
	idt_interrupt(_irq6, VECTOR_M+6);
	irq_on(6);
}

u_int32_t fdc_probe(void)
{
	u_int8_t v;
	
	fdc_reset();

	/* Detect controller */
	fdc_sendbyte(CMD_VERSION);
	v=fdc_getbyte();

	if ( v == 0x80 ) {
		kprint(_FDC_PREFIX "NEC765 floppy controller\n");
	}else{
		kprint(_FDC_PREFIX "Enhanced floppy controller (0x%x)\n", v);
	}
	
	return 1;
}

u_int32_t fdc_setup(void)
{
	return 1;
}

int32_t fdc_blk_rw(int32_t write, u_int32_t blk, char *buf, u_int32_t count)
{
	if ( write == RWB_READ ) {
		fdc_read(blk, buf, count);
		return 0;
	}else{
		kprint(_FDC_PREFIX "Floppy driver is read only!!\n");
		return -ERDONLY;
	}
}

int fdc_read(u_int32_t block, char *buf, int len)
{
	int head, track, sector;
	int tries, i;
	
	/* Convert block number in to physical addressing */
	fdc_linear2hts(block, &head, &track, &sector);

	for(tries=0; tries<3; tries++)
	{
		if ( inb(FDC_DIR) & 0x80 ) {
			fdc_seek(1);
			fdc_recalibrate();
			kprint(_FDC_PREFIX "Disk changed on read!\n");
			break;
		}

		if ( !fdc_seek(track) ) {
			kprint(_FDC_PREFIX "Error seeking to track %u!\n", track);
			fdc_reset();
			break;
		}
		
		/* Set data rate (500k/sec) */
		outb(FDC_CCR, 0);

		/* Do DMA transfer */
		dma_read(2, tb, len*512);
		fdc_sendbyte(CMD_READ);
		fdc_sendbyte(head<<2);
		fdc_sendbyte(track);
		fdc_sendbyte(head);
		fdc_sendbyte(sector);
		fdc_sendbyte(2);
		fdc_sendbyte(geometry->spt);
		fdc_sendbyte(geometry->g3_rw);
		fdc_sendbyte(0xff);

		if ( !fdc_wait(1) ) {
			kprint(_FDC_PREFIX "Timeout while reading!\n");
			return 0;
		}

		if ( (status[0] & 0xc0) == 0 ) break;

		kprint(_FDC_PREFIX "Read failed, trying again\n");
		fdc_recalibrate();
	}if ( tries==3 ) return 0;


	for (i=0; i<(len*512); i++) {
		*((char *)(((u_int32_t)buf)+i))=*((char *)(((u_int32_t)tb)+i));
	}
	
	return 1;
}
