#include <kernel.h>
#include <arch/keyboard.h>
#include <arch/8259a.h>
#include <arch/idt.h>

static volatile u_int8_t kb_buffer[255];
static volatile u_int8_t kb_len=0;

u_int8_t led_mask=0x00;
u_int8_t modifiers=0x00;

/* Keymap for an PC/AT keyboard */
kb_scan_t keymap[]={
	{0x02, '1',	'!'},
	{0x03, '2',	'\"'},
	{0x04, '3',	''},
	{0x05, '4',	'$'},
	{0x06, '5',	'%'},
	{0x07, '6',	'^'},
	{0x08, '7',	'&'},
	{0x09, '8',	'*'},
	{0x0a, '9',	'('},
	{0x0b, '0',	')'},
	{0x0c, '-',	'_'},
	{0x0d, '=',	'+'},
	{0x0e, '\x8',	'\x8'},	/* <- Backspace */
	{0x0f, '\t',	'\t'},	/* Tab */
	{0x10, 'q',	'Q'},
	{0x11, 'w',	'W'},
	{0x12, 'e',	'E'},
	{0x13, 'r',	'R'},
	{0x14, 't',	'T'},
	{0x15, 'y',	'Y'},
	{0x16, 'u',	'U'},
	{0x17, 'i',	'I'},
	{0x18, 'o',	'O'},
	{0x19, 'p',	'P'},
	{0x1a, '[',	'{'},
	{0x1b, ']',	'}'},
	{0x1c, '\n',	'\n'},	/* Return */
	{0x1e, 'a',	'A'},
	{0x1f, 's',	'S'},
	{0x20, 'd',	'D'},
	{0x21, 'f',	'F'},
	{0x22, 'g',	'G'},
	{0x23, 'h',	'H'},
	{0x24, 'j',	'J'},
	{0x25, 'k',	'K'},
	{0x26, 'l',	'L'},
	{0x27, ';',	':'},
	{0x28, '\'',	'@'},
	{0x29, '`',	''},
	{0x2b, '\\',	'|'},
	{0x2c, 'z',	'Z'},
	{0x2d, 'x',	'X'},
	{0x2e, 'c',	'C'},
	{0x2f, 'v',	'V'},
	{0x30, 'b',	'B'},
	{0x31, 'n',	'N'},
	{0x32, 'm',	'M'},
	{0x33, ',',	'<'},
	{0x34, '.',	'>'},
	{0x35, '/',	'?'},
	{0x39, ' ',	' '},
};

void kb_isr(void)
{
	u_int8_t x= inb(0x60);
	
	if ( kb_len == 255 ) {
		kprint("Keyboard buffer overflow. key=0x%x\n", x);
		irq_eoi(1);
		return;
	}
	
	kb_buffer[kb_len]=x;
	kb_len++;	
	irq_eoi(1);
	return;
}
void _irq1(void);
asm(
	".globl _irq1		\n"
	"_irq1:			\n"
	"	pusha		\n"
	"	call kb_isr	\n"
	"	popa		\n"
	"	iret		\n"
);

void __init keyboard_init(void)
{
	int x;
	idt_interrupt(_irq1, VECTOR_M+1);

	/* Blank the keyboard buffer */
	for (x=0; x<256; x++) {
		kb_buffer[x]=0;
	}
	
}

u_int32_t keyboard_probe(void)
{
	return 1;
}

u_int32_t keyboard_setup(void)
{
	/* Unmask the keyboard IRQ, and blank the leds */
	kb_setleds();
	irq_on(1);
	return 1;
}

void kb_led_on(u_int8_t led)
{
	led_mask |= led;
	kb_setleds();
}

void kb_led_off(u_int8_t led)
{
	led_mask &= ~led;
	kb_setleds();
}

void kb_setleds(void)
{
	outb(0x60, 0xED);
	while(inb(0x64) & 2);
	outb(0x60, led_mask);     
	while(inb(0x64) & 2);
}

u_int8_t kb_scan2ascii(u_int8_t scan)
{
	u_int32_t i;

	switch ( scan ) {
		case 0x2a:
		case 0x36:
			MASK_SET(modifiers, KB_MOD_SHIFT);
			return 0;
		case 0xaa:
		case 0xb6:
			MASK_UNSET(modifiers, KB_MOD_SHIFT);
			return 0;
		case 0x3a:
			if ( MASK(modifiers, KB_MOD_CAPS) ) {
				MASK_UNSET(modifiers, KB_MOD_CAPS);
				kb_led_off(KB_LED_CAP);
			}else{
				MASK_SET(modifiers, KB_MOD_CAPS);
				kb_led_on(KB_LED_CAP);
			}
				
			return 0;
		default:
	}
	
	for (i=0; i<sizeof(keymap)/sizeof(kb_scan_t); i++ ) {
		if ( keymap[i].scan == scan ) {
			if ( modifiers == 0)
				return keymap[i].ascii;
		
			if ( modifiers == (1<<KB_MOD_SHIFT) )
				return keymap[i].shift;
		}
	}
	
	//kprint("Unknown PC/AT scancode: 0x%x\n", scan);
	return 0;
}

u_int32_t kb_keypress(void)
{
	return kb_len;
}

u_int8_t kb_readkey(void)
{
	int x;
	u_int8_t k;
	
	if ( !kb_len ) return 0;
	
	/* Mask off the interrupt while we play with the bufffer */
	irq_off(1);

	k=kb_buffer[0];
	for (x=1; x<kb_len+1; x++){
		kb_buffer[x-1]=kb_buffer[x];
	}

	kb_len--;
	
	/* Reenable the interrupt ASAP */
	irq_on(1);

	return kb_scan2ascii(k);
}
